__all__ = ("TdeLc", "TdePopMetric", "TdePopMetricQuality", "generate_tde_pop_slicer")

import glob
import os

import numpy as np
from rubin_scheduler.data import get_data_dir
from rubin_scheduler.utils import survey_start_mjd, uniform_sphere

import rubin_sim.maf.metrics as metrics
import rubin_sim.maf.slicers as slicers
from rubin_sim.phot_utils import DustValues


class TdeLc:
    """
    Read in some TDE lightcurves

    Parameters
    ----------
    file_list : list of str (None)
        List of file paths to load. If None, loads up all the files from data/tde/
    """

    def __init__(self, file_list=None):
        if file_list is None:
            sims_maf_contrib_dir = get_data_dir()
            file_list = glob.glob(os.path.join(sims_maf_contrib_dir, "maf/tde/*.dat"))

        lcs = []
        for filename in file_list:
            lcs.append(np.genfromtxt(filename, dtype=[("ph", "f8"), ("mag", "f8"), ("filter", "U1")]))

        # Let's organize the data in to a list of dicts for easy lookup
        self.data = []
        filternames = "ugrizy"
        for lc in lcs:
            new_dict = {}
            for filtername in filternames:
                infilt = np.where(lc["filter"] == filtername)
                new_dict[filtername] = lc[infilt]
            self.data.append(new_dict)

    def interp(self, t, filtername, lc_indx=0):
        """
        t : array of floats
            The times to interpolate the light curve to.
        filtername : str
            The filter. one of ugrizy
        lc_index : int (0)
            Which file to use.
        """

        result = np.interp(
            t,
            self.data[lc_indx][filtername]["ph"],
            self.data[lc_indx][filtername]["mag"],
            left=99,
            right=99,
        )
        return result


class TdePopMetric(metrics.BaseMetric):
    def __init__(
        self,
        metric_name="TDEsPopMetric",
        mjd_col="observationStartMJD",
        m5_col="fiveSigmaDepth",
        filter_col="filter",
        night_col="night",
        pts_needed=2,
        file_list=None,
        mjd0=None,
        **kwargs,
    ):
        maps = ["DustMap"]
        self.mjd_col = mjd_col
        self.m5_col = m5_col
        self.filter_col = filter_col
        self.night_col = night_col
        self.pts_needed = pts_needed

        self.lightcurves = TdeLc(file_list=file_list)
        self.mjd0 = survey_start_mjd() if mjd0 is None else mjd0

        dust_properties = DustValues()
        self.ax1 = dust_properties.ax1

        cols = [self.mjd_col, self.m5_col, self.filter_col, self.night_col]
        super(TdePopMetric, self).__init__(
            col=cols, units="Detected, 0 or 1", metric_name=metric_name, maps=maps, **kwargs
        )

    def _pre_peak_detect(self, data_slice, slice_point, mags, t):
        """
        Simple detection criteria
        """
        result = 0
        # Simple alert criteria. Could make more in depth, or use reduce functions
        # to have multiple criteria checked.
        pre_peak_detected = np.where((t < 0) & (mags < data_slice[self.m5_col]))[0]

        if pre_peak_detected.size > self.pts_needed:
            result = 1
        return result

    def _some_color_detect(self, data_slice, slice_point, mags, t):
        result = 1
        # 1 detection pre peak
        pre_peak_detected = np.where((t < -10) & (mags < data_slice[self.m5_col]))[0]
        if np.size(pre_peak_detected) < 1:
            return 0

        # At least 3 filters within 10 days of peak
        around_peak = np.where((np.abs(t) < 5) & (mags < data_slice[self.m5_col]))[0]
        if np.size(np.unique(data_slice[self.filter_col][around_peak])) < 3:
            return 0

        # At least 2 bands after peak
        post_peak = np.where((t > 10) & (t < 30) & (mags < data_slice[self.m5_col]))[0]
        if np.size(np.unique(data_slice[self.filter_col][post_peak])) < 2:
            return 0

        return result

    def _some_color_pu_detect(self, data_slice, slice_point, mags, t):
        result = 1
        # 1 detection pre peak
        pre_peak_detected = np.where((t < -10) & (mags < data_slice[self.m5_col]))[0]
        if np.size(pre_peak_detected) < 1:
            return 0

        # 1 detection in u and any other band near peak
        around_peak = np.where((np.abs(t) < 5) & (mags < data_slice[self.m5_col]))[0]
        filters = np.unique(data_slice[self.filter_col][around_peak])
        if np.size(filters) < 2:
            return 0
        if "u" not in filters:
            return 0

        post_peak = np.where((t > 10) & (t < 30) & (mags < data_slice[self.m5_col]))[0]
        filters = np.unique(data_slice[self.filter_col][post_peak])
        if np.size(filters) < 2:
            return 0
        if "u" not in filters:
            return 0

        return result

    def run(self, data_slice, slice_point=None):
        result = {}
        t = data_slice[self.mjd_col] - self.mjd0 - slice_point["peak_time"]
        mags = np.zeros(t.size, dtype=float)

        for filtername in np.unique(data_slice[self.filter_col]):
            infilt = np.where(data_slice[self.filter_col] == filtername)
            mags[infilt] = self.lightcurves.interp(t[infilt], filtername, lc_indx=slice_point["file_indx"])
            # Apply dust extinction on the light curve
            mags[infilt] += self.ax1[filtername] * slice_point["ebv"]

        result["pre_peak"] = self._pre_peak_detect(data_slice, slice_point, mags, t)
        result["some_color"] = self._some_color_detect(data_slice, slice_point, mags, t)
        result["some_color_pu"] = self._some_color_pu_detect(data_slice, slice_point, mags, t)

        return result

    def reduce_prepeak(self, metric):
        return metric["pre_peak"]

    def reduce_some_color(self, metric):
        return metric["some_color"]

    def reduce_some_color_pu(self, metric):
        return metric["some_color_pu"]


class TdePopMetricQuality(metrics.BaseMetric):
    def __init__(
        self,
        metric_name="TDEsPopMetricQuality",
        mjd_col="observationStartMJD",
        m5_col="fiveSigmaDepth",
        filter_col="filter",
        night_col="night",
        tmin=-30,
        tmax=100,
        file_list=None,
        mjd0=59853.5,
        **kwargs,
    ):
        maps = ["DustMap"]
        self.mjd_col = mjd_col
        self.m5_col = m5_col
        self.filter_col = filter_col
        self.night_col = night_col
        self.tmin = tmin
        self.tmax = tmax

        self.lightcurves = TdeLc(file_list=file_list)
        self.mjd0 = mjd0

        dust_properties = DustValues()
        self.ax1 = dust_properties.ax1

        cols = [self.mjd_col, self.m5_col, self.filter_col, self.night_col]
        super(TdePopMetricQuality, self).__init__(
            col=cols,
            units="Bad: 0, Good (obs every 2 night): 1",
            metric_name=metric_name,
            maps=maps,
            **kwargs,
        )

    def _some_color_pnum_detect(self, data_slice, slice_point, mags, t):
        # 1 detection pre peak
        pre_peak_detected = np.where((t < -10) & (mags < data_slice[self.m5_col]))[0]
        if np.size(pre_peak_detected) < 1:
            return 0

        # At least 3 filters within 10 days of peak
        around_peak = np.where((np.abs(t) < 5) & (mags < data_slice[self.m5_col]))[0]
        if np.size(np.unique(data_slice[self.filter_col][around_peak])) < 3:
            return 0

        # At least 2 bands after peak
        post_peak = np.where((t > 10) & (t < 30) & (mags < data_slice[self.m5_col]))[0]
        if np.size(np.unique(data_slice[self.filter_col][post_peak])) < 2:
            return 0

        # count number of data points in the light curve
        obs_points = np.where((t > self.tmin) & (t < self.tmax) & (mags < data_slice[self.m5_col]))[0]

        # define the time range around peak in which the number of data points is measured
        t_range = self.tmax - self.tmin

        # number of data points / time range gives a "score" for light curve quality
        # 0: did not pass some_color requirements;
        # 1: passed some_color requirements and has 1 data point every other night
        nresult = np.size(obs_points) / t_range

        return nresult

    def _some_color_pu_pnum_detect(self, data_slice, slice_point, mags, t):
        # 1 detection pre peak
        pre_peak_detected = np.where((t < -10) & (mags < data_slice[self.m5_col]))[0]
        if np.size(pre_peak_detected) < 1:
            return 0

        # 1 detection in u and any other band near peak
        around_peak = np.where((np.abs(t) < 5) & (mags < data_slice[self.m5_col]))[0]
        filters = np.unique(data_slice[self.filter_col][around_peak])
        if np.size(filters) < 2:
            return 0
        if "u" not in filters:
            return 0

        # 1 detecion in u and any other band post peak
        post_peak = np.where((t > 10) & (t < 30) & (mags < data_slice[self.m5_col]))[0]
        filters = np.unique(data_slice[self.filter_col][post_peak])
        if np.size(filters) < 2:
            return 0
        if "u" not in filters:
            return 0

        # count number of data points in the light curve
        obs_points = np.where((t > self.tmin) & (t < self.tmax) & (mags < data_slice[self.m5_col]))[0]

        # define the time range around peak in which the number of data points is measured
        t_range = self.tmax - self.tmin

        # number of data points / time range gives a "score" for light curve quality
        # 0: did not pass some_color_pu requirements;
        # 1: passed some_color_pu requirements and has 1 data point every other night
        nresult = np.size(obs_points) / t_range

        return nresult

    def run(self, data_slice, slice_point=None):
        result = {}
        t = data_slice[self.mjd_col] - self.mjd0 - slice_point["peak_time"]
        mags = np.zeros(t.size, dtype=float)

        for filtername in np.unique(data_slice[self.filter_col]):
            infilt = np.where(data_slice[self.filter_col] == filtername)
            mags[infilt] = self.lightcurves.interp(t[infilt], filtername, lc_indx=slice_point["file_indx"])
            # Apply dust extinction on the light curve
            mags[infilt] += self.ax1[filtername] * slice_point["ebv"]

        result["some_color_pnum"] = self._some_color_pnum_detect(data_slice, slice_point, mags, t)
        result["some_color_pu_pnum"] = self._some_color_pu_pnum_detect(data_slice, slice_point, mags, t)

        return result

    def reduce_some_color_pnum(self, metric):
        return metric["some_color_pnum"]

    def reduce_some_color_pu_pnum(self, metric):
        return metric["some_color_pu_pnum"]


def generate_tde_pop_slicer(t_start=1, t_end=3652, n_events=10000, seed=42, n_files=7):
    """Generate a population of TDE events, and put the info about them into a UserPointSlicer object

    Parameters
    ----------
    t_start : float (1)
        The night to start tde events on (days)
    t_end : float (3652)
        The final night of TDE events
    n_events : int (10000)
        The number of TDE events to generate
    seed : float
        The seed passed to np.random
    n_files : int (7)
        The number of different TDE lightcurves to use
    """

    ra, dec = uniform_sphere(n_events, seed=seed)
    peak_times = np.random.uniform(low=t_start, high=t_end, size=n_events)
    file_indx = np.floor(np.random.uniform(low=0, high=n_files, size=n_events)).astype(int)

    # Set up the slicer to evaluate the catalog we just made
    slicer = slicers.UserPointsSlicer(ra, dec, lat_lon_deg=True, badval=0)
    # Add any additional information about each object to the slicer
    slicer.slice_points["peak_time"] = peak_times
    slicer.slice_points["file_indx"] = file_indx
    return slicer



#####################################################################################################
# Purpose: calculate the coadded 5-sigma depth from various survey strategies. Incudes functionality
# to consider various survey strategies, mask shallow borders, create/save/show relevant plots, do
# an alm analysis, and save data.

__all__ = ("coadd_m5_analysis",)

import copy
import os

import healpy as hp
import matplotlib.pyplot as plt

# Humna Awan: humna.awan@rutgers.edu
#####################################################################################################
import numpy as np
from matplotlib.ticker import FuncFormatter, MaxNLocator

import rubin_sim.maf.db as db
import rubin_sim.maf.maps as maps
import rubin_sim.maf.metric_bundles as metricBundles
import rubin_sim.maf.metrics as metrics
import rubin_sim.maf.plots as plots
import rubin_sim.maf.slicers as slicers
import rubin_sim.maf.stackers as mafStackers  # stackers in sims_maf
from rubin_sim.maf.maf_contrib.lss_obs_strategy.alm_plots import alm_plots
from rubin_sim.maf.maf_contrib.lss_obs_strategy.constants_for_pipeline import plot_color
from rubin_sim.maf.maf_contrib.lss_obs_strategy.masking_algorithm_generalized import (
    masking_algorithm_generalized,
)


def coadd_m5_analysis(
    path,
    dbfile,
    run_name,
    slair=False,
    wf_dand_dd_fs=False,
    no_dith_only=False,
    best_dith_only=False,
    some_dith_only=False,
    specified_dith=None,
    nside=128,
    filter_band="r",
    include_dust_extinction=False,
    saveun_masked_coadd_data=False,
    pixel_radius_for_masking=5,
    cut_off_year=None,
    plot_skymap=True,
    plot_cartview=True,
    unmasked_color_min=None,
    unmasked_color_max=None,
    masked_color_min=None,
    masked_color_max=None,
    n_ticks=5,
    plot_power_spectrum=True,
    show_plots=True,
    save_figs=True,
    alm_analysis=True,
    ra_range=[-50, 50],
    dec_range=[-65, 5],
    save_masked_coadd_data=True,
):
    """

    Analyze the artifacts induced in the coadded 5sigma depth due to imperfect observing strategy.
      - Creates an output directory for subdirectories containing the specified things to save.
      - Creates, shows, and saves comparison plots.
      - Returns the metricBundle object containing the calculated coadded depth, and the output directory name.

    Parameters
    ----------
    path : str
        path to the main directory where output directory is to be saved.
    dbfile : str
        path to the OpSim output file, e.g. to a copy of enigma_1189
    run_name : str
        run name tag to identify the output of specified OpSim output, e.g. 'enigma1189'
    slair : `bool`
        set to True if analysis on a SLAIR output.
        Default: False
    wf_dand_dd_fs : `bool`
        set to True if want to consider both WFD survet and DDFs. Otherwise will only work
        with WFD. Default: False
    no_dith_only : `bool`
        set to True if only want to consider the undithered survey. Default: False
    best_dith_only : `bool`
        set to True if only want to consider RandomDitherFieldPerVisit.
        Default: False
    some_dith_only : `bool`
        set to True if only want to consider undithered and a few dithered surveys.
        Default: False
    specified_dith : str
        specific dither strategy to run.
        Default: None
    nside : int
        HEALpix resolution parameter. Default: 128
    filter_band : str
        any one of 'u', 'g', 'r', 'i', 'z', 'y'. Default: 'r'
    include_dust_extinction : `bool`
        set to include dust extinction. Default: False
    saveun_masked_coadd_data : `bool`
        set to True to save data before border masking. Default: False
    pixel_radius_for_masking : int
        number of pixels to mask along the shallow border. Default: 5
    cut_off_year : int
        year cut to restrict analysis to only a subset of the survey.
        Must range from 1 to 9, or None for the full survey analysis (10 yrs).
        Default: None
    plot_skymap : `bool`
        set to True if want to plot skymaps. Default: True
    plot_cartview : `bool`
        set to True if want to plot cartview plots. Default: False
    unmasked_color_min : float
        lower limit on the colorscale for unmasked skymaps. Default: None
    unmasked_color_max : float
        upper limit on the colorscale for unmasked skymaps. Default: None
    masked_color_min : float
        lower limit on the colorscale for border-masked skymaps. Default: None
    masked_color_max : float
        upper limit on the colorscale for border-masked skymaps. Default: None
    n_ticks : int
        (number of ticks - 1) on the skymap colorbar. Default: 5
    plot_power_spectrum : `bool`
        set to True if want to plot powerspectra. Default: True
    show_plots : `bool`
        set to True if want to show figures. Default: True
    save_figs : `bool`
        set to True if want to save figures. Default: True
    alm_analysis : `bool`
        set to True to perform the alm analysis. Default: True
    ra_range : float array
        range of right ascention (in degrees) to consider in alm  cartview plot;
        applicable when alm_analysis=True. Default: [-50,50]
    dec_range : float array
        range of declination (in degrees) to consider in alm cartview plot;
        applicable when alm_analysis=True. Default: [-65,5]
    save_masked_coadd_data : `bool`
        set to True to save the coadded depth data after the border
        masking. Default: True
    """
    # ------------------------------------------------------------------------
    # read in the database
    if slair:
        # slair database
        opsdb = db.Database(dbfile, defaultTable="observations")
    else:
        # OpSim database
        opsdb = db.OpsimDatabase(dbfile)

    # ------------------------------------------------------------------------
    # set up the out_dir
    zeropt_tag = ""
    if cut_off_year is not None:
        zeropt_tag = "%syearCut" % cut_off_year
    else:
        zeropt_tag = "fullSurveyPeriod"

    if include_dust_extinction:
        dust_tag = "withDustExtinction"
    else:
        dust_tag = "noDustExtinction"

    region_type = ""
    if wf_dand_dd_fs:
        region_type = "WFDandDDFs_"

    out_dir = "coaddM5Analysis_%snside%s_%s_%spixelRadiusForMasking_%sBand_%s_%s_directory" % (
        region_type,
        nside,
        dust_tag,
        pixel_radius_for_masking,
        filter_band,
        run_name,
        zeropt_tag,
    )
    print("# out_dir: %s" % out_dir)
    results_db = db.ResultsDb(out_dir=out_dir)

    # ------------------------------------------------------------------------
    # set up the sql constraint
    if wf_dand_dd_fs:
        if cut_off_year is not None:
            night_cut_off = (cut_off_year) * 365.25
            sqlconstraint = 'night<=%s and filter=="%s"' % (night_cut_off, filter_band)
        else:
            sqlconstraint = 'filter=="%s"' % filter_band
    else:
        # set up the propID and units on the ra, dec
        if slair:  # no prop ID; only WFD is simulated.
            wfd_where = ""
            ra_dec_in_deg = True
        else:
            prop_ids, prop_tags = opsdb.fetchPropInfo()
            wfd_where = "%s and " % opsdb.createSQLWhere("WFD", prop_tags)
            ra_dec_in_deg = opsdb.raDecInDeg
        # set up the year cutoff
        if cut_off_year is not None:
            night_cut_off = (cut_off_year) * 365.25
            sqlconstraint = '%snight<=%s and filter=="%s"' % (
                wfd_where,
                night_cut_off,
                filter_band,
            )
        else:
            sqlconstraint = '%sfilter=="%s"' % (wfd_where, filter_band)
    print("# sqlconstraint: %s" % sqlconstraint)

    # ------------------------------------------------------------------------
    # setup all the slicers
    slicer = {}
    stacker_list = {}

    if (
        specified_dith is not None
    ):  # would like to add all the stackers first and then keep only the one that is specified
        best_dith_only, no_dith_only = False, False

    if best_dith_only:
        stacker_list["RandomDitherFieldPerVisit"] = [
            mafStackers.RandomDitherFieldPerVisitStacker(degrees=ra_dec_in_deg, random_seed=1000)
        ]
        slicer["RandomDitherFieldPerVisit"] = slicers.HealpixSlicer(
            lon_col="randomDitherFieldPerVisitRa",
            lat_col="randomDitherFieldPerVisitDec",
            lat_lon_deg=ra_dec_in_deg,
            nside=nside,
            use_cache=False,
        )
    else:
        if slair:
            slicer["NoDither"] = slicers.HealpixSlicer(
                lon_col="RA",
                lat_col="dec",
                lat_lon_deg=ra_dec_in_deg,
                nside=nside,
                use_cache=False,
            )
        else:
            slicer["NoDither"] = slicers.HealpixSlicer(
                lon_col="fieldRA",
                lat_col="fieldDec",
                lat_lon_deg=ra_dec_in_deg,
                nside=nside,
                use_cache=False,
            )
        if some_dith_only and not no_dith_only:
            # stacker_list['RepulsiveRandomDitherFieldPerVisit'] = [myStackers.RepulsiveRandomDitherFieldPerVisitStacker(degrees=ra_dec_in_deg,
            #                                                                                                          random_seed=1000)]
            # slicer['RepulsiveRandomDitherFieldPerVisit'] = slicers.HealpixSlicer(lonCol='repulsiveRandomDitherFieldPerVisitRa',
            #                                                                    latCol='repulsiveRandomDitherFieldPerVisitDec',
            #                                                                    latLonDeg=ra_dec_in_deg, nside=nside,
            #                                                                    use_cache=False)
            slicer["SequentialHexDitherFieldPerNight"] = slicers.HealpixSlicer(
                lon_col="hexDitherFieldPerNightRa",
                lat_col="hexDitherFieldPerNightDec",
                lat_lon_deg=ra_dec_in_deg,
                nside=nside,
                use_cache=False,
            )
            slicer["PentagonDitherPerSeason"] = slicers.HealpixSlicer(
                lon_col="pentagonDitherPerSeasonRa",
                lat_col="pentagonDitherPerSeasonDec",
                lat_lon_deg=ra_dec_in_deg,
                nside=nside,
                use_cache=False,
            )
        elif not no_dith_only:
            # random dithers on different timescales
            stacker_list["RandomDitherPerNight"] = [
                mafStackers.RandomDitherPerNightStacker(degrees=ra_dec_in_deg, random_seed=1000)
            ]
            stacker_list["RandomDitherFieldPerNight"] = [
                mafStackers.RandomDitherFieldPerNightStacker(degrees=ra_dec_in_deg, random_seed=1000)
            ]
            stacker_list["RandomDitherFieldPerVisit"] = [
                mafStackers.RandomDitherFieldPerVisitStacker(degrees=ra_dec_in_deg, random_seed=1000)
            ]

            # rep random dithers on different timescales
            # stacker_list['RepulsiveRandomDitherPerNight'] = [myStackers.RepulsiveRandomDitherPerNightStacker(degrees=ra_dec_in_deg,
            #                                                                                                random_seed=1000)]
            # stacker_list['RepulsiveRandomDitherFieldPerNight'] = [myStackers.RepulsiveRandomDitherFieldPerNightStacker(degrees=ra_dec_in_deg,
            #                                                                                                          random_seed=1000)]
            # stacker_list['RepulsiveRandomDitherFieldPerVisit'] = [myStackers.RepulsiveRandomDitherFieldPerVisitStacker(degrees=ra_dec_in_deg,
            #                                                                                                          random_seed=1000)]
            # set up slicers for different dithers
            # random dithers on different timescales
            slicer["RandomDitherPerNight"] = slicers.HealpixSlicer(
                lon_col="randomDitherPerNightRa",
                lat_col="randomDitherPerNightDec",
                lat_lon_deg=ra_dec_in_deg,
                nside=nside,
                use_cache=False,
            )
            slicer["RandomDitherFieldPerNight"] = slicers.HealpixSlicer(
                lon_col="randomDitherFieldPerNightRa",
                lat_col="randomDitherFieldPerNightDec",
                lat_lon_deg=ra_dec_in_deg,
                nside=nside,
                use_cache=False,
            )
            slicer["RandomDitherFieldPerVisit"] = slicers.HealpixSlicer(
                lon_col="randomDitherFieldPerVisitRa",
                lat_col="randomDitherFieldPerVisitDec",
                lat_lon_deg=ra_dec_in_deg,
                nside=nside,
                use_cache=False,
            )
            # rep random dithers on different timescales
            # slicer['RepulsiveRandomDitherPerNight'] = slicers.HealpixSlicer(lonCol='repulsiveRandomDitherPerNightRa',
            #                                                               latCol='repulsiveRandomDitherPerNightDec',
            #                                                               latLonDeg=ra_dec_in_deg, nside=nside, use_cache=False)
            # slicer['RepulsiveRandomDitherFieldPerNight'] = slicers.HealpixSlicer(lonCol='repulsiveRandomDitherFieldPerNightRa',
            #                                                                    latCol='repulsiveRandomDitherFieldPerNightDec',
            #                                                                    latLonDeg=ra_dec_in_deg, nside=nside,
            #                                                                    use_cache=False)
            # slicer['RepulsiveRandomDitherFieldPerVisit'] = slicers.HealpixSlicer(lonCol='repulsiveRandomDitherFieldPerVisitRa',
            #                                                                    latCol='repulsiveRandomDitherFieldPerVisitDec',
            #                                                                    latLonDeg=ra_dec_in_deg, nside=nside,
            #                                                                    use_cache=False)
            # spiral dithers on different timescales
            slicer["FermatSpiralDitherPerNight"] = slicers.HealpixSlicer(
                lon_col="fermatSpiralDitherPerNightRa",
                lat_col="fermatSpiralDitherPerNightDec",
                lat_lon_deg=ra_dec_in_deg,
                nside=nside,
                use_cache=False,
            )
            slicer["FermatSpiralDitherFieldPerNight"] = slicers.HealpixSlicer(
                lon_col="fermatSpiralDitherFieldPerNightRa",
                lat_col="fermatSpiralDitherFieldPerNightDec",
                lat_lon_deg=ra_dec_in_deg,
                nside=nside,
                use_cache=False,
            )
            slicer["FermatSpiralDitherFieldPerVisit"] = slicers.HealpixSlicer(
                lon_col="fermatSpiralDitherFieldPerVisitRa",
                lat_col="fermatSpiralDitherFieldPerVisitDec",
                lat_lon_deg=ra_dec_in_deg,
                nside=nside,
                use_cache=False,
            )
            # hex dithers on different timescales
            slicer["SequentialHexDitherPerNight"] = slicers.HealpixSlicer(
                lon_col="hexDitherPerNightRa",
                lat_col="hexDitherPerNightDec",
                lat_lon_deg=ra_dec_in_deg,
                nside=nside,
                use_cache=False,
            )
            slicer["SequentialHexDitherFieldPerNight"] = slicers.HealpixSlicer(
                lon_col="hexDitherFieldPerNightRa",
                lat_col="hexDitherFieldPerNightDec",
                lat_lon_deg=ra_dec_in_deg,
                nside=nside,
                use_cache=False,
            )
            slicer["SequentialHexDitherFieldPerVisit"] = slicers.HealpixSlicer(
                lon_col="hexDitherFieldPerVisitRa",
                lat_col="hexDitherFieldPerVisitDec",
                lat_lon_deg=ra_dec_in_deg,
                nside=nside,
                use_cache=False,
            )
            # per season dithers
            slicer["PentagonDitherPerSeason"] = slicers.HealpixSlicer(
                lon_col="pentagonDitherPerSeasonRa",
                lat_col="pentagonDitherPerSeasonDec",
                lat_lon_deg=ra_dec_in_deg,
                nside=nside,
                use_cache=False,
            )
            slicer["PentagonDiamondDitherPerSeason"] = slicers.HealpixSlicer(
                lon_col="pentagonDiamondDitherPerSeasonRa",
                lat_col="pentagonDiamondDitherPerSeasonDec",
                lat_lon_deg=ra_dec_in_deg,
                nside=nside,
                use_cache=False,
            )
            slicer["SpiralDitherPerSeason"] = slicers.HealpixSlicer(
                lon_col="spiralDitherPerSeasonRa",
                lat_col="spiralDitherPerSeasonDec",
                lat_lon_deg=ra_dec_in_deg,
                nside=nside,
                use_cache=False,
            )
    if specified_dith is not None:
        stacker_list_, slicer_ = {}, {}
        if specified_dith in slicer.keys():
            if specified_dith.__contains__(
                "Random"
            ):  # only Random dithers have a stacker object for rand seed specification
                stacker_list_[specified_dith] = stacker_list[specified_dith]
            slicer_[specified_dith] = slicer[specified_dith]
        else:
            raise ValueError(
                "Invalid value for specified_dith: %s. Allowed values include one of the following:\n%s"
                % (specified_dith, slicer.keys())
            )
        stacker_list, slicer = stacker_list_, slicer_

    # ------------------------------------------------------------------------
    if slair:
        m5_col = "fivesigmadepth"
    else:
        m5_col = "fiveSigmaDepth"
    # set up the metric
    if include_dust_extinction:
        # include dust extinction when calculating the co-added depth
        coadd_metric = metrics.ExgalM5(m5_col=m5_col, lsstFilter=filter_band)
    else:
        coadd_metric = metrics.Coaddm5Metric(m5col=m5col)
    dust_map = maps.DustMap(
        interp=False, nside=nside
    )  # include dust_map; actual in/exclusion of dust is handled by the galaxyCountMetric

    # ------------------------------------------------------------------------
    # set up the bundle
    coadd_bundle = {}
    for dither in slicer:
        if dither in stacker_list:
            coadd_bundle[dither] = metricBundles.MetricBundle(
                coadd_metric,
                slicer[dither],
                sqlconstraint,
                stacker_list=stacker_list[dither],
                run_name=run_name,
                metadata=dither,
                maps_list=[dust_map],
            )
        else:
            coadd_bundle[dither] = metricBundles.MetricBundle(
                coadd_metric,
                slicer[dither],
                sqlconstraint,
                run_name=run_name,
                metadata=dither,
                maps_list=[dust_map],
            )

    # ------------------------------------------------------------------------
    # run the analysis
    if include_dust_extinction:
        print("\n# Running coadd_bundle with dust extinction ...")
    else:
        print("\n# Running coadd_bundle without dust extinction ...")
    c_group = metricBundles.MetricBundleGroup(
        coadd_bundle, opsdb, out_dir=out_dir, results_db=results_db, save_early=False
    )
    c_group.run_all()

    # ------------------------------------------------------------------------
    plot_handler = plots.PlotHandler(out_dir=out_dir, results_db=results_db, thumbnail=False, savefig=False)

    print("# Number of pixels in the survey region (before masking the border):")
    for dither in coadd_bundle:
        print(
            "  %s: %s"
            % (
                dither,
                len(np.where(coadd_bundle[dither].metricValues.mask == False)[0]),
            )
        )

    # ------------------------------------------------------------------------
    # save the unmasked data?
    if saveun_masked_coadd_data:
        out_dir_new = "unmaskedCoaddData"
        for b in coadd_bundle:
            coadd_bundle[b].write(out_dir=out_dir_new)

    # ------------------------------------------------------------------------
    # mask the edges
    print("\n# Masking the edges for coadd ...")
    coadd_bundle = masking_algorithm_generalized(
        coadd_bundle,
        plot_handler,
        data_label="$%s$-band Coadded Depth" % filter_band,
        nside=nside,
        pixel_radius=pixel_radius_for_masking,
        plot_intermediate_plots=False,
        plot_final_plots=False,
        print_final_info=True,
    )
    # ------------------------------------------------------------------------
    # Calculate total power
    summarymetric = metrics.TotalPowerMetric()
    for dither in coadd_bundle:
        coadd_bundle[dither].set_summary_metrics(summarymetric)
        coadd_bundle[dither].compute_summary_stats()
        print(
            "# Total power for %s case is %f." % (dither, coadd_bundle[dither].summary_values["TotalPower"])
        )
    print("")

    # ------------------------------------------------------------------------
    # run the alm analysis
    if alm_analysis:
        alm_plots(
            path,
            out_dir,
            copy.deepcopy(coadd_bundle),
            nside=nside,
            filterband=filter_band,
            ra_range=ra_range,
            dec_range=dec_range,
            show_plots=show_plots,
        )
    # ------------------------------------------------------------------------
    # save the masked data?
    if save_masked_coadd_data and (pixel_radius_for_masking > 0):
        out_dir_new = "maskedCoaddData"
        for b in coadd_bundle:
            coadd_bundle[b].write(out_dir=out_dir_new)

    # ------------------------------------------------------------------------
    # plot comparison plots
    if len(coadd_bundle.keys()) > 1:  # more than one key
        # set up the directory
        out_dir_comp = "coaddM5ComparisonPlots"
        if not os.path.exists("%s%s/%s" % (path, out_dir, out_dir_comp)):
            os.makedirs("%s%s/%s" % (path, out_dir, out_dir_comp))
        # ------------------------------------------------------------------------
        # plot for the power spectra
        cl = {}
        for dither in plot_color:
            if dither in coadd_bundle:
                cl[dither] = hp.anafast(
                    hp.remove_dipole(
                        coadd_bundle[dither].metricValues.filled(coadd_bundle[dither].slicer.badval)
                    ),
                    lmax=500,
                )
                ell = np.arange(np.size(cl[dither]))
                plt.plot(
                    ell,
                    (cl[dither] * ell * (ell + 1)) / 2.0 / np.pi,
                    color=plot_color[dither],
                    linestyle="-",
                    label=dither,
                )
        plt.xlabel(r"$\ell$")
        plt.ylabel(r"$\ell(\ell+1)C_\ell/(2\pi)$")
        plt.xlim(0, 500)
        fig = plt.gcf()
        fig.set_size_inches(12.5, 10.5)
        leg = plt.legend(labelspacing=0.001)
        for legobj in leg.legendHandles:
            legobj.set_linewidth(4.0)
        filename = "powerspectrum_comparison_all.png"
        plt.savefig(
            "%s%s/%s/%s" % (path, out_dir, out_dir_comp, filename),
            bbox_inches="tight",
            format="png",
        )
        plt.show()

        # create the histogram
        scale = hp.nside2pixarea(nside, degrees=True)

        def tick_formatter(y, pos):
            return "%d" % (y * scale)  # convert pixel count to area

        bin_size = 0.01
        for dither in plot_color:
            if dither in coadd_bundle:
                ind = np.where(coadd_bundle[dither].metricValues.mask == False)[0]
                bin_all = int(
                    (
                        max(coadd_bundle[dither].metricValues.data[ind])
                        - min(coadd_bundle[dither].metricValues.data[ind])
                    )
                    / bin_size
                )
                plt.hist(
                    coadd_bundle[dither].metricValues.data[ind],
                    bins=bin_all,
                    label=dither,
                    histtype="step",
                    color=plot_color[dither],
                )
        ax = plt.gca()
        ymin, ymax = ax.get_ylim()
        n_yticks = 10.0
        wanted_y_max = ymax * scale
        wanted_y_max = 10.0 * np.ceil(float(wanted_y_max) / 10.0)
        increment = 5.0 * np.ceil(float(wanted_y_max / n_yticks) / 5.0)
        wanted_array = np.arange(0, wanted_y_max, increment)
        ax.yaxis.set_ticks(wanted_array / scale)
        ax.yaxis.set_major_formatter(FuncFormatter(tick_formatter))
        plt.xlabel("$%s$-band Coadded Depth" % filter_band)
        plt.ylabel("Area (deg$^2$)")
        fig = plt.gcf()
        fig.set_size_inches(12.5, 10.5)
        leg = plt.legend(labelspacing=0.001, loc=2)
        for legobj in leg.legendHandles:
            legobj.set_linewidth(2.0)
        filename = "histogram_comparison.png"
        plt.savefig(
            "%s%s/%s/%s" % (path, out_dir, out_dir_comp, filename),
            bbox_inches="tight",
            format="png",
        )
        plt.show()
        # ------------------------------------------------------------------------
        # plot power spectra for the separte panel
        tot_keys = len(list(coadd_bundle.keys()))
        if tot_keys > 1:
            plt.clf()
            n_cols = 2
            n_rows = int(np.ceil(float(tot_keys) / n_cols))
            fig, ax = plt.subplots(n_rows, n_cols)
            plot_row = 0
            plot_col = 0
            for dither in list(plot_color.keys()):
                if dither in list(coadd_bundle.keys()):
                    ell = np.arange(np.size(cl[dither]))
                    ax[plot_row, plot_col].plot(
                        ell,
                        (cl[dither] * ell * (ell + 1)) / 2.0 / np.pi,
                        color=plot_color[dither],
                        label=dither,
                    )
                    if plot_row == n_rows - 1:
                        ax[plot_row, plot_col].set_xlabel(r"$\ell$")
                    ax[plot_row, plot_col].set_ylabel(r"$\ell(\ell+1)C_\ell/(2\pi)$")
                    ax[plot_row, plot_col].yaxis.set_major_locator(MaxNLocator(3))
                    if dither != "NoDither":
                        ax[plot_row, plot_col].set_ylim(0, 0.0035)
                    ax[plot_row, plot_col].set_xlim(0, 500)
                    plot_row += 1
                    if plot_row > n_rows - 1:
                        plot_row = 0
                        plot_col += 1
            fig.set_size_inches(20, int(n_rows * 30 / 7.0))
            filename = "powerspectrum_sepPanels.png"
            plt.savefig(
                "%s%s/%s/%s" % (path, out_dir, out_dir_comp, filename),
                bbox_inches="tight",
                format="png",
            )
            plt.show()
    return coadd_bundle, out_dir



"""Evaluate some bulk properties of the sky coverage
"""
__all__ = ("meanRADec", "eastWestBias")

import numpy as np

import rubin_sim.maf.metric_bundles as mb
import rubin_sim.maf.metrics as metrics
import rubin_sim.maf.slicers as slicers

from .col_map_dict import col_map_dict


def meanRADec(colmap=None, runName="opsim", extraSql=None, extraInfoLabel=None):
    """Plot the range of RA/Dec as a function of night.

    Parameters
    ----------
    colmap : `dict`, optional
        A dictionary with a mapping of column names.
    runName : `str`, optional
        The name of the simulated survey.
    extraSql : `str`, optional
        Additional constraint to add to any sql constraints (e.g. 'night<365')
    extraInfoLabel : `str`, optional
        Additional info_label to add before any below (i.e. "WFD").

    Returns
    -------
    metric_bundleDict : `dict` of `maf.MetricBundle`
    """
    if colmap is None:
        colmap = col_map_dict()
    bundleList = []

    group = "RA Dec coverage"

    subgroup = "All visits"
    if extraInfoLabel is not None:
        subgroup = extraInfoLabel

    displayDict = {"group": group, "subgroup": subgroup, "order": 0}

    ra_metrics = [
        metrics.MeanAngleMetric(colmap["ra"]),
        metrics.FullRangeAngleMetric(colmap["ra"]),
    ]
    dec_metrics = [
        metrics.MeanMetric(colmap["dec"]),
        metrics.MinMetric(colmap["dec"]),
        metrics.MaxMetric(colmap["dec"]),
    ]
    for m in ra_metrics:
        slicer = slicers.OneDSlicer(slice_col_name=colmap["night"], bin_size=1)
        if not colmap["raDecDeg"]:
            plotDict = {"y_min": np.radians(-5), "y_max": np.radians(365)}
        else:
            plotDict = {"y_min": -5, "y_max": 365}
        bundle = mb.MetricBundle(
            m,
            slicer,
            extraSql,
            info_label=extraInfoLabel,
            display_dict=displayDict,
            plot_dict=plotDict,
        )
        bundleList.append(bundle)

    for m in dec_metrics:
        slicer = slicers.OneDSlicer(slice_col_name=colmap["night"], bin_size=1)
        bundle = mb.MetricBundle(m, slicer, extraSql, info_label=extraInfoLabel, display_dict=displayDict)
        bundleList.append(bundle)

    # Set the run_name for all bundles and return the bundleDict.
    for b in bundleList:
        b.set_run_name(runName)
    return mb.make_bundles_dict_from_list(bundleList)


def eastWestBias(colmap=None, runName="opsim", extraSql=None, extraInfoLabel=None):
    """Plot the number of observations to the east vs to the west, per night.

    Parameters
    ----------
    colmap : `dict`, optional
        A dictionary with a mapping of column names.
    runName : `str`, optional
        The name of the simulated survey.
    extraSql : `str`, optional
        Additional constraint to add to any sql constraints (e.g. 'night<365')
    extraInfoLabel : `str`, optional
        Additional info_label to add before any below (i.e. "WFD").

    Returns
    -------
    metric_bundleDict : `dict` of `maf.MetricBundle`
    """
    if colmap is None:
        colmap = col_map_dict()
    bundleList = []

    group = "East vs West"

    subgroup = "All visits"
    if extraInfoLabel is not None:
        subgroup = extraInfoLabel

    displayDict = {"group": group, "subgroup": subgroup, "order": 0}

    eastvswest = 180
    if not colmap["raDecDeg"]:
        eastvswest = np.radians(eastvswest)

    displayDict["caption"] = "Number of visits per night that occur with azimuth <= 180."
    if extraSql is not None:
        displayDict["caption"] += " With additional sql constraint %s." % extraSql
    metric = metrics.CountMetric(colmap["night"], metric_name="Nvisits East")
    slicer = slicers.OneDSlicer(slice_col_name=colmap["night"], bin_size=1)
    sql = "%s <= %f" % (colmap["az"], eastvswest)
    if extraSql is not None:
        sql = "(%s) and (%s)" % (sql, extraSql)
    plotDict = {"color": "orange", "label": "East"}
    bundle = mb.MetricBundle(
        metric,
        slicer,
        sql,
        info_label=extraInfoLabel,
        display_dict=displayDict,
        plot_dict=plotDict,
    )
    bundleList.append(bundle)

    displayDict["caption"] = "Number of visits per night that occur with azimuth > 180."
    if extraSql is not None:
        displayDict["caption"] += " With additional sql constraint %s." % extraSql
    metric = metrics.CountMetric(colmap["night"], metric_name="Nvisits West")
    slicer = slicers.OneDSlicer(slice_col_name=colmap["night"], bin_size=1)
    sql = "%s > %f" % (colmap["az"], eastvswest)
    if extraSql is not None:
        sql = "(%s) and (%s)" % (sql, extraSql)
    plotDict = {"color": "blue", "label": "West"}
    bundle = mb.MetricBundle(
        metric,
        slicer,
        sql,
        info_label=extraInfoLabel,
        display_dict=displayDict,
        plot_dict=plotDict,
    )
    bundleList.append(bundle)

    # Set the run_name for all bundles and return the bundleDict.
    for b in bundleList:
        b.set_run_name(runName)
    return mb.make_bundles_dict_from_list(bundleList)



__all__ = ("TwoDSubsetData", "OneDSubsetData")


import matplotlib.pyplot as plt
import numpy as np
from matplotlib import colors

from .perceptual_rainbow import make_pr_cmap
from .plot_handler import BasePlotter

perceptual_rainbow = make_pr_cmap()


class TwoDSubsetData(BasePlotter):
    """
    Plot 2 axes from the slicer.sliceColList, identified by
    plot_dict['xaxis']/['yaxis'], given the metric_values at all
    slicepoints [sums over non-visible axes].
    """

    def __init__(self):
        self.plot_type = "2DBinnedData"
        self.object_plotter = False
        self.default_plot_dict = {
            "title": None,
            "xlabel": None,
            "ylable": None,
            "units": None,
            "log_scale": False,
            "clims": None,
            "cmap": perceptual_rainbow,
            "cbar_format": None,
        }

    def __call__(self, metric_values, slicer, user_plot_dict, fignum=None):
        """
        Parameters
        ----------
        metricValue : numpy.ma.MaskedArray
        slicer : rubin_sim.maf.slicers.NDSlicer
        user_plot_dict: dict
            Dictionary of plot parameters set by user (overrides default values).
            'xaxis' and 'yaxis' values define which axes of the nd data to plot along the x/y axes.
        fignum : int
            Matplotlib figure number to use (default = None, starts new figure).

        Returns
        -------
        int
           Matplotlib figure number used to create the plot.
        """
        if slicer.slicer_name != "NDSlicer":
            raise ValueError("TwoDSubsetData plots ndSlicer metric values")
        fig = plt.figure(fignum)
        plot_dict = {}
        plot_dict.update(self.default_plot_dict)
        plot_dict.update(user_plot_dict)
        if "xaxis" not in plot_dict or "yaxis" not in plot_dict:
            raise ValueError("xaxis and yaxis must be specified in plot_dict")
        xaxis = plot_dict["xaxis"]
        yaxis = plot_dict["yaxis"]
        # Reshape the metric data so we can isolate the values to plot
        # (just new view of data, not copy).
        newshape = []
        for b in slicer.bins:
            newshape.append(len(b) - 1)
        newshape.reverse()
        md = metric_values.reshape(newshape)
        # Sum over other dimensions. Note that masked values are not included in sum.
        sumaxes = list(range(slicer.nD))
        sumaxes.remove(xaxis)
        sumaxes.remove(yaxis)
        sumaxes = tuple(sumaxes)
        md = md.sum(sumaxes)
        # Plot the histogrammed data.
        # Plot data.
        x, y = np.meshgrid(slicer.bins[xaxis][:-1], slicer.bins[yaxis][:-1])
        if plot_dict["log_scale"]:
            norm = colors.LogNorm()
        else:
            norm = None
        if plot_dict["clims"] is None:
            im = plt.contourf(x, y, md, 250, norm=norm, extend="both", cmap=plot_dict["cmap"])
        else:
            im = plt.contourf(
                x,
                y,
                md,
                250,
                norm=norm,
                extend="both",
                cmap=plot_dict["cmap"],
                vmin=plot_dict["clims"][0],
                vmax=plot_dict["clims"][1],
            )
        xlabel = plot_dict["xlabel"]
        if xlabel is None:
            xlabel = slicer.sliceColList[xaxis]
        plt.xlabel(xlabel)
        ylabel = plot_dict["ylabel"]
        if ylabel is None:
            ylabel = slicer.sliceColList[yaxis]
        plt.ylabel(ylabel)
        cb = plt.colorbar(
            im,
            aspect=25,
            extend="both",
            orientation="horizontal",
            format=plot_dict["cbar_format"],
        )
        cb.set_label(plot_dict["units"])
        plt.title(plot_dict["title"])
        return fig.number


class OneDSubsetData(BasePlotter):
    """
    Plot a single axes from the sliceColList, identified by plot_dict['axis'],
    given the metric_values at all slicepoints [sums over non-visible axes].
    """

    def __init__(self):
        self.plot_type = "1DBinnedData"
        self.object_plotter = False
        self.default_plot_dict = {
            "title": None,
            "xlabel": None,
            "ylabel": None,
            "label": None,
            "units": None,
            "log_scale": False,
            "histRange": None,
            "filled": False,
            "alpha": 0.5,
            "cmap": perceptual_rainbow,
            "cbar_format": None,
        }

    def plot_binned_data1_d(self, metric_values, slicer, user_plot_dict, fignum=None):
        """
        Parameters
        ----------
        metricValue : numpy.ma.MaskedArray
        slicer : rubin_sim.maf.slicers.NDSlicer
        user_plot_dict: dict
            Dictionary of plot parameters set by user (overrides default values).
            'axis' keyword identifies which axis to show in the plot (along xaxis of plot).
        fignum : int
            Matplotlib figure number to use (default = None, starts new figure).

        Returns
        -------
        int
           Matplotlib figure number used to create the plot.
        """
        if slicer.slicer_name != "NDSlicer":
            raise ValueError("TwoDSubsetData plots ndSlicer metric values")
        fig = plt.figure(fignum)
        plot_dict = {}
        plot_dict.update(self.default_plot_dict)
        plot_dict.update(user_plot_dict)
        if "axis" not in plot_dict:
            raise ValueError("axis for 1-d plot must be specified in plot_dict")
        # Reshape the metric data so we can isolate the values to plot
        # (just new view of data, not copy).
        axis = plot_dict["axis"]
        newshape = []
        for b in slicer.bins:
            newshape.append(len(b) - 1)
        newshape.reverse()
        md = metric_values.reshape(newshape)
        # Sum over other dimensions. Note that masked values are not included in sum.
        sumaxes = list(range(slicer.nD))
        sumaxes.remove(axis)
        sumaxes = tuple(sumaxes)
        md = md.sum(sumaxes)
        # Plot the histogrammed data.
        leftedge = slicer.bins[axis][:-1]
        width = np.diff(slicer.bins[axis])
        if plot_dict["filled"]:
            plt.bar(
                leftedge,
                md,
                width,
                label=plot_dict["label"],
                linewidth=0,
                alpha=plot_dict["alpha"],
                log=plot_dict["log_scale"],
            )
        else:
            x = np.ravel(list(zip(leftedge, leftedge + width)))
            y = np.ravel(list(zip(md, md)))
            if plot_dict["log_scale"]:
                plt.semilogy(x, y, label=plot_dict["label"])
            else:
                plt.plot(x, y, label=plot_dict["label"])
        plt.ylabel(plot_dict["ylabel"])
        xlabel = plot_dict["xlabel"]
        if xlabel is None:
            xlabel = slicer.sliceColName[axis]
            if plot_dict["units"] is not None:
                xlabel += " (" + plot_dict["units"] + ")"
        plt.xlabel(xlabel)
        if plot_dict["histRange"] is not None:
            plt.xlim(plot_dict["histRange"])
        plt.title(plot_dict["title"])
        return fig.number



__all__ = ("XrbLc", "XRBPopMetric", "generate_xrb_pop_slicer")

import numpy as np
from rubin_scheduler.data import get_data_dir
from rubin_scheduler.utils import survey_start_mjd
from scipy.stats import loguniform

from rubin_sim.maf.utils import m52snr
from rubin_sim.phot_utils import DustValues

from ..metrics import BaseMetric
from ..slicers import UserPointsSlicer


class XrbLc:
    """Synthesize XRB outburst lightcurves."""

    def __init__(self, seed=42):
        dust_properties = DustValues()
        self.ax1 = dust_properties.ax1

        self.rng = np.random.default_rng(seed)

    def lmxb_abs_mags(self, size=1):
        """Return LMXB absolute magnitudes per LSST filter.

        Absolute magnitude relation is taken from Casares 2018 (2018MNRAS.473.5195C)
        Colors are taken from M. Johnson+ 2019 (2019MNRAS.484...19J)

        Parameters
        ----------
        size : `int`
            Number of samples to generate.

        Returns
        -------
        abs_mags : `list` [`dict`]
            Absolute magnitudes for each LSST filter.
        Porbs: `array` [`float`]
            Randomized orbital periods in days
        """

        # Derive random orbital periods from the sample in Casares 18 Table 4
        # Since there are significant outliers from a single Gaussian sample,
        # take random choices with replacement and then perturb them fractionally
        catalog__porb = np.array(
            [
                33.85,
                6.4713,
                2.5445,
                1.7557,
                1.5420,
                0.5213,
                0.4326,
                0.3441,
                0.3230,
                0.3205,
                0.2852,
                0.2740,
                0.2122,
                0.1699,
                0.1352,
                0.117,
                0.1006,
            ]
        )

        sample__porbs = self.rng.choice(catalog__porb, size=size)
        sample__porbs *= self.rng.uniform(low=0.5, high=1.5, size=size)

        # lmxb_abs_mag_r = 4.6 # johnson+18
        # Casares 18
        lmxb_abs_mags_r = 4.64 - 3.69 * np.log10(sample__porbs)

        return [
            {
                "u": lmxb_abs_mag_r + 4.14,
                "g": lmxb_abs_mag_r + 3.24,
                "r": lmxb_abs_mag_r,
                "i": lmxb_abs_mag_r + 0.33,
                "z": lmxb_abs_mag_r + 1.05,
                "y": lmxb_abs_mag_r + 2.36,
            }
            for lmxb_abs_mag_r in lmxb_abs_mags_r
        ], sample__porbs

    def outburst_params(self, size=1):
        """Return a parameters at random characterizing the outburst.

        Uses distributions from Chen, Shrader, & Livio 1997 (ApJ 491, 312).

        Returns
        -------
        params : `list` [`dict`]
            Rise, decay, and amplitude parameters.
        """

        # rise timescale (Fig. 11): 50% 1-2 day lognormal,
        #                           50% loguniform 0.5-50 days

        flip = self.rng.uniform(size=size)
        w = flip > 0.5
        tau_rise = np.zeros(size) * np.nan
        tau_rise[w] = loguniform.rvs(0.5, 50, size=np.sum(w))
        tau_rise[~w] = self.rng.lognormal(0, 0.2, size=np.sum(~w)) + 0.3

        # amplitude = rng.lognormal(9,1,size=1)[0]
        # delta_mag = -2.5 * np.log10(amplitude)
        # The Chen+ 10**4 X-ray flux amplitudes imply -10 optical delta mags,
        # which are generally larger than observed.
        # We'll use a fixed -6 delta mag for simplicity
        delta_mag = -6
        amplitude = 10 ** (-0.4 * delta_mag) * np.ones(size)

        tau_decay = self.rng.lognormal(2.9, 0.65, size=size)

        duration = (tau_rise + tau_decay) + np.log(amplitude)

        abs_mags, porbs = self.lmxb_abs_mags(size=size)

        return [
            {
                "tau_rise": tr,
                "tau_decay": td,
                "amplitude": amp,
                "outburst_duration": dur,
                "abs_mag": abs_mag,
                "orbital_period": Porb,
            }
            for (tr, td, amp, dur, abs_mag, Porb) in zip(
                tau_rise, tau_decay, amplitude, duration, abs_mags, porbs
            )
        ]

    def fred(self, t, amplitude, tau_rise, tau_decay):
        """Fast-rise, exponential decay function.

        Amplitude is defined at the peak time = sqrt(tau_rise*tau_decay).

        See e.g., Tarnopolski 2021 for discussion.

        Parameters
        ----------
        t : `array` [`float`]
            The times relative to the start of the outburst
        amplitude : `float`
            Peak amplitude
        tau_rise : `float`
            E-folding time for the rise
        tau_decay : `float`
            E-folding time for the decay
        """
        return amplitude * np.exp(2 * np.sqrt(tau_rise / tau_decay)) * np.exp(-tau_rise / t - t / tau_decay)

    def lightcurve(self, t, filtername, params):
        """Generate an XRB outburst lightcurve for given times and a single filter.

        Uses a simple fast-rise, exponential decay with parameters taken from
        Chen, Shrader, & Livio 1997 (ApJ 491, 312).

        For now we ignore the late time linear decay (Tetarenko+2018a,b, and
        references therein.)

        Parameters
        ----------
        t : `array` [`float`]
            The times relative to the start of the outburst
        filtername : `str`
            The filter. one of ugrizy
        params : `dict`
            parameters for the FRED lightcurve.

        Returns
        -------
        lc : `array`
            Magnitudes of the outburst at the specified times in the given filter
        """

        # fill lightcurve with nondetections
        lc = np.ones(len(t)) * 99

        # FRED
        woutburst = (t >= 0) & (t <= params["outburst_duration"])
        if np.sum(woutburst):
            lc[woutburst] = params["abs_mag"][filtername] + -2.5 * np.log10(
                self.fred(
                    t[woutburst],
                    params["amplitude"],
                    params["tau_rise"],
                    params["tau_decay"],
                )
            )

        return lc

    def detectable_duration(self, params, ebv, distance):
        """Determine time range an outburst is detectable with perfect sampling.

        Does not consider visibility constraints.

        Parameters
        ----------
        params : `dict`
            lightcurve parameters for XrbLc
        ebv : `float`
            E(B-V)
        distance : `float`
            distance in kpc

        Returns
        ----------
        visible_start_time : `float`
            first time relative to outburst start that the outburst could be detected
        visible_end_time : `float`
            last time relative to outburst start that the outburst could be detected
        """

        nmodelt = 10000
        t = np.linspace(0, params["outburst_duration"], nmodelt)

        lsst_single_epoch_depth = {
            "u": 23.9,
            "g": 25.0,
            "r": 24.7,
            "i": 24.0,
            "z": 23.3,
            "y": 22.1,
        }

        visible_start_time = np.inf
        visible_end_time = -np.inf

        for filtername in ["u", "g", "r", "i", "z", "y"]:
            mags = self.lightcurve(t, filtername, params)

            # Apply dust extinction on the light curve
            a_x = self.ax1[filtername] * ebv
            mags += a_x
            distmod = 5 * np.log10(distance * 1.0e3) - 5.0
            mags += distmod

            wdetectable = np.where(mags <= lsst_single_epoch_depth[filtername])[0]
            if len(wdetectable) == 0:
                continue

            if t[wdetectable[0]] < visible_start_time:
                visible_start_time = t[wdetectable[0]]

            if t[wdetectable[-1]] >= visible_end_time:
                visible_end_time = t[wdetectable[-1]]

        if np.isinf(visible_start_time):
            return np.nan, np.nan

        return visible_start_time, visible_end_time


class XRBPopMetric(BaseMetric):
    def __init__(
        self,
        metric_name="XRBPopMetric",
        mjd_col="observationStartMJD",
        m5_col="fiveSigmaDepth",
        filter_col="filter",
        night_col="night",
        pts_needed=2,
        mjd0=None,
        output_lc=False,
        badval=-666,
        **kwargs,
    ):
        # maps = ["DustMap"]
        self.mjd_col = mjd_col
        self.m5_col = m5_col
        self.filter_col = filter_col
        self.night_col = night_col
        self.pts_needed = pts_needed
        # `bool` variable, if True the light curve will be exported
        self.output_lc = output_lc

        self.lightcurves = XrbLc()
        self.mjd0 = survey_start_mjd() if mjd0 is None else mjd0

        dust_properties = DustValues()
        self.ax1 = dust_properties.ax1

        cols = [self.mjd_col, self.m5_col, self.filter_col, self.night_col]
        super(XRBPopMetric, self).__init__(
            col=cols,
            units="Detected, 0 or 1",
            metric_name=metric_name,
            # maps=maps,
            badval=badval,
            **kwargs,
        )
        self.comment = "Number or characterization of XRBs."

    def _ever_detect(self, where_detected):
        """Simple detection criteria: detect at least a certain number of times"""
        # Detected data points
        return np.size(where_detected) >= self.pts_needed

    def _number_of_detections(self, where_detected):
        """Count total number of detections."""
        return len(where_detected)

    def _early_detect(self, where_detected, time, early_window_days=7.0, n_early_detections=2):
        """Detection near the start of the outburst.

        Parameters
        ----------
        where_detected : `array`
            indexes corresponding to 5 sigma detections
        mags : `array`
            magnitudes obtained interpolating models on the data_slice
        time : `array`
            relative times
        early_window_days : `float`
            time since start of outburst
        n_early_detections : `int`
            number of required early detections
        """

        return np.sum(time[where_detected] <= early_window_days) >= n_early_detections

    def _mean_time_between_detections(self, t):
        """Calculate the mean time between detections over the visible interval.

        Parameters
        ----------
        t : `array`
            Times of detections, bracketed by the start and end visibility times

        Return
        ----------
        med_dt : `float`
             separation between observations.
        """

        return np.mean(np.sort(np.diff(t)))

    def _possible_to_detect(self, visible_duration):
        """Return True if the outburst is ever bright enough for LSST to detect

        Parameters
        ----------
        visible_duration : `float`
            Length of time the outburst is above LSST's fiducial limiting mag.
            May be nan.

        Return
        ----------
        detectable : `bool`
             Return True if outburst is ever detectable by LSST.
        """

        return ~np.isnan(visible_duration)

    def run(self, data_slice, slice_point=None):
        result = {}
        t = data_slice[self.mjd_col] - self.mjd0 - slice_point["start_time"]
        mags = np.zeros(t.size, dtype=float)

        for filtername in np.unique(data_slice[self.filter_col]):
            infilt = np.where(data_slice[self.filter_col] == filtername)
            mags[infilt] = self.lightcurves.lightcurve(t[infilt], filtername, slice_point["outburst_params"])
            # Apply dust extinction on the light curve
            a_x = self.ax1[filtername] * slice_point["ebv"]
            mags[infilt] += a_x

            distmod = 5 * np.log10(slice_point["distance"] * 1.0e3) - 5.0
            mags[infilt] += distmod

        # Find the detected points
        where_detected = np.where((mags < data_slice[self.m5_col]))[0]
        # Magnitude uncertainties with Gaussian approximation
        snr = m52snr(mags, data_slice[self.m5_col])
        mags_unc = 2.5 * np.log10(1.0 + 1.0 / snr)

        result["possible_to_detect"] = self._possible_to_detect(slice_point["visible_duration"])
        result["ever_detect"] = self._ever_detect(where_detected)
        result["early_detect"] = self._early_detect(where_detected, t)
        result["number_of_detections"] = self._number_of_detections(where_detected)

        if result["number_of_detections"] > 1:
            result["mean_time_between_detections"] = self._mean_time_between_detections(
                [
                    slice_point["visible_start_time"],
                    *t[where_detected].tolist(),
                    slice_point["visible_end_time"],
                ]
            )
        else:
            result["mean_time_between_detections"] = -999

        # Export the light curve
        if self.output_lc is True:
            wdet = mags < data_slice[self.m5_col]
            mags[~wdet] = 99.0
            result["lc"] = [
                data_slice[self.mjd_col],
                mags,
                mags_unc,
                data_slice[self.m5_col],
                data_slice[self.filter_col],
            ]
            result["lc_colnames"] = ("t", "mag", "mag_unc", "maglim", "filter")

        return result

    def reduce_possible_to_detect(self, metric):
        return metric["possible_to_detect"]

    def reduce_ever_detect(self, metric):
        return metric["ever_detect"]

    def reduce_early_detect(self, metric):
        return metric["early_detect"]

    def reduce_number_of_detections(self, metric):
        return metric["number_of_detections"]

    def reduce_mean_time_between_detections(self, metric):
        tt = metric["mean_time_between_detections"]
        if tt < 0:
            return self.badval
        else:
            return tt


def generate_xrb_pop_slicer(t_start=1, t_end=3652, n_events=10000, seed=42):
    """Generate a population of XRB events, and put the info about them
    into a UserPointSlicer object

    Parameters
    ----------
    t_start : `float`
        The night to start an XRB outburst on (days; default 1)
    t_end : `float`
        The final night of XRBs events (days; default 3652)
    n_events : `int`
        The number of XRB outbursts to generate (default 10000)
    seed : `float`
        The seed passed to np.random (default 42)
    """

    rng = np.random.default_rng(seed)

    datadir = get_data_dir()
    xrb_sample = np.genfromtxt(datadir + "/maf/xrb/sample_xrb_positions.csv.gz", delimiter=",")

    nsamples, nfields = xrb_sample.shape

    xrb_lc_gen = XrbLc()

    # select a random subsample
    event_idxs = rng.choice(np.arange(nsamples), size=n_events, replace=True)
    ra = xrb_sample[event_idxs, 0]
    dec = xrb_sample[event_idxs, 1]
    distance_kpc = xrb_sample[event_idxs, 2]
    ebv = xrb_sample[event_idxs, 3]

    start_times = rng.uniform(low=t_start, high=t_end, size=n_events)

    # Set up the slicer to evaluate the catalog we just made
    slicer = UserPointsSlicer(ra, dec, lat_lon_deg=True, badval=0)
    # Add any additional information about each object to the slicer
    slicer.slice_points["start_time"] = start_times
    slicer.slice_points["distance"] = distance_kpc
    # use our own 3-d dust map extinctions
    slicer.slice_points["ebv"] = ebv
    # generate random parameters for this event
    slicer.slice_points["outburst_params"] = xrb_lc_gen.outburst_params(size=n_events)

    # determine detectable durations
    visible_starts = []
    visible_ends = []
    visible_durations = []
    for idx, param in enumerate(slicer.slice_points["outburst_params"]):
        start, end = xrb_lc_gen.detectable_duration(param, ebv[idx], distance_kpc[idx])
        visible_starts.append(start)
        visible_ends.append(end)
        visible_durations.append(end - start)

    slicer.slice_points["visible_start_time"] = visible_starts
    slicer.slice_points["visible_end_time"] = visible_ends
    slicer.slice_points["visible_duration"] = visible_durations

    return slicer



#!/usr/bin/env python

# Mike Lund - Vanderbilt University
# mike.lund@gmail.com
# Last edited 8/15/2015
# Description: Takes a given set of galactic coordinates and a stellar mass range, then calculates the number of stars within that range that will be fainter than mag 16, and have sufficiently low noise in the given band. For use with Field Star Count metric
import math
import sys

import numpy as np
from scipy.optimize import fsolve, minimize, newton

from . import abs_mag, coords, spec_type, stellardensity
from .starcount import starcount

xi = 1.0
alpha = 2.35


def IMF(lower, upper):
    exp = alpha - 1.0
    part1 = xi / exp
    part2 = lower**-exp - upper**-exp
    part3 = part1 * part2
    return part3


def get_distance(apparent, absolute):
    part1 = (apparent - absolute) / 5.0
    part2 = 10 * 10**part1
    return part2


def noise_opt(m, band, sigma):
    part1 = mag_error(m, band)
    part2 = part1**2.0
    part3 = sigma**2.0
    total = part2 - part3
    return total


def mag_error(ap_mag, band, calcm5=0):  # apparent magnitude and band
    gamma = {"u": 0.037, "g": 0.038, "r": 0.039, "i": 0.039, "z": 0.040, "y": 0.040}
    m5 = {"u": 23.9, "g": 25.0, "r": 24.7, "i": 24.0, "z": 23.3, "y": 22.1}
    if calcm5 == 0:
        calcm5 = m5[band]
    X = 10.0 ** (0.4 * (ap_mag - calcm5))
    random_2 = np.sqrt((0.04 - gamma[band]) * X + gamma[band] * X * X)
    error_2 = random_2**2.0 + (0.005) ** 2.0
    mag_error = np.sqrt(error_2)
    return mag_error


def noise_calc(band):
    gamma = {"u": 0.037, "g": 0.038, "r": 0.039, "i": 0.039, "z": 0.040, "y": 0.040}
    m5 = {"u": 23.9, "g": 25.0, "r": 24.7, "i": 24.0, "z": 23.3, "y": 22.1}
    sigma = 0.03
    sigma_sys = 0.005
    fun = (
        lambda x: sigma_sys**2
        + (0.04 - gamma[band]) * 10 ** (0.4 * (x - m5[band]))
        + gamma[band] * 10 ** (0.8 * (x - m5[band]))
        - sigma**2
    )
    return newton(fun, 25)


def dist_calc(mass, band):
    # mass to spectral type
    # spectral type to absolute mag
    bands = ["z", "y", "i", "r", "g", "u"]
    output = abs_mag.abs_mag(spec_type.spec_type(mass))[0]
    indexvalue = bands.index(band)
    absolutemag = output[indexvalue][0]
    apparent = noise_calc(band)
    dist_min = get_distance(16, absolutemag)
    dist_max = get_distance(apparent, absolutemag)
    return dist_min, dist_max
    # abs mag to apparent mag ranges, > 16, noise dependent upper limit


def starcount_bymass(eq_ra, eq_dec, m1, m2, band):
    masses = np.linspace(m1, m2, num=20)
    totmass = IMF(m1, m2)
    totmass = IMF(0.2, 1.04)
    massbins = IMF(masses[:-1], masses[1:])
    massfractions = massbins / totmass
    distances = [dist_calc(x, band) for x in masses[:-1]]
    starcounts = [y * starcount(eq_ra, eq_dec, x[0], x[1]) for x, y in zip(distances, massfractions)]
    return sum(starcounts)


if __name__ == "__main__":
    print(
        starcount_bymass(
            float(sys.argv[1]),
            float(sys.argv[2]),
            float(sys.argv[3]),
            float(sys.argv[4]),
            sys.argv[5],
        )
    )



__all__ = ("CrowdingM5Metric", "CrowdingMagUncertMetric", "NstarsMetric")

import healpy as hp
import numpy as np
from scipy.interpolate import interp1d

from rubin_sim.maf.metrics import BaseMetric

# Modifying from Knut Olson's fork at:
# https://github.com/knutago/sims_maf_contrib/blob/master/tutorials/CrowdingMetric.ipynb


def _comp_crowd_error(mag_vector, lum_func, seeing, single_mag=None):
    """
    Compute the photometric crowding error given the luminosity function and best seeing.

    Parameters
    ----------
    mag_vector : np.array
        Stellar magnitudes.
    lum_func : np.array
        Stellar luminosity function.
    seeing : float
        The best seeing conditions. Assuming forced-photometry can use the best seeing conditions
        to help with confusion errors.
    single_mag : float (None)
        If single_mag is None, the crowding error is calculated for each mag in mag_vector. If
        single_mag is a float, the crowding error is interpolated to that single value.

    Returns
    -------
    np.array
        Magnitude uncertainties.

    Equation from Olsen, Blum, & Rigaut 2003, AJ, 126, 452
    """
    lum_area_arcsec = 3600.0**2
    lum_vector = 10 ** (-0.4 * mag_vector)
    coeff = np.sqrt(np.pi / lum_area_arcsec) * seeing / 2.0
    my_int = (np.add.accumulate((lum_vector**2 * lum_func)[::-1]))[::-1]
    temp = np.sqrt(my_int) / lum_vector
    if single_mag is not None:
        interp = interp1d(mag_vector, temp)
        temp = interp(single_mag)
    crowd_error = coeff * temp
    return crowd_error


class CrowdingM5Metric(BaseMetric):
    """Return the magnitude at which the photometric error exceeds crowding_error threshold."""

    def __init__(
        self,
        crowding_error=0.1,
        filtername="r",
        seeing_col="seeingFwhmGeom",
        metric_name=None,
        maps=["StellarDensityMap"],
        **kwargs,
    ):
        """
        Parameters
        ----------
        crowding_error : float, optional
            The magnitude uncertainty from crowding in magnitudes. Default 0.1 mags.
        filtername: str, optional
            The bandpass in which to calculate the crowding limit. Default r.
        seeing_col : str, optional
            The name of the seeing column.
        m5Col : str, optional
            The name of the m5 depth column.
        maps : list of str, optional
            Names of maps required for the metric.

        Returns
        -------
        float
        The magnitude of a star which has a photometric error of `crowding_error`
        """

        cols = [seeing_col]
        units = "mag"
        self.crowding_error = crowding_error
        self.filtername = filtername
        self.seeing_col = seeing_col
        if metric_name is None:
            metric_name = "Crowding to Precision %.2f" % (crowding_error)
        super().__init__(col=cols, maps=maps, units=units, metric_name=metric_name, **kwargs)

    def run(self, data_slice, slice_point=None):
        # Set mag_vector to the same length as starLumFunc (lower edge of mag bins)
        mag_vector = slice_point[f"starMapBins_{self.filtername}"][1:]
        # Pull up density of stars at this point in the sky
        lum_func = slice_point[f"starLumFunc_{self.filtername}"]
        # Calculate the crowding error using the best seeing value (in any filter?)
        crowd_error = _comp_crowd_error(mag_vector, lum_func, seeing=min(data_slice[self.seeing_col]))
        # Locate at which point crowding error is greater than user-defined limit
        above_crowd = np.where(crowd_error >= self.crowding_error)[0]

        if np.size(above_crowd) == 0:
            result = max(mag_vector)
        else:
            crowd_mag = mag_vector[max(above_crowd[0] - 1, 0)]
            result = crowd_mag

        return result


class NstarsMetric(BaseMetric):
    """Return the number of stars visible above some uncertainty limit,
    taking image depth and crowding into account.
    """

    def __init__(
        self,
        crowding_error=0.1,
        filtername="r",
        seeing_col="seeingFwhmGeom",
        m5_col="fiveSigmaDepth",
        metric_name=None,
        maps=["StellarDensityMap"],
        ignore_crowding=False,
        **kwargs,
    ):
        """
        Parameters
        ----------
        crowding_error : float, optional
            The magnitude uncertainty from crowding in magnitudes. Default 0.1 mags.
        filtername: str, optional
            The bandpass in which to calculate the crowding limit. Default r.
        seeing_col : str, optional
            The name of the seeing column.
        m5_col : str, optional
            The name of the m5 depth column.
        maps : list of str, optional
            Names of maps required for the metric.
        ignore_crowding : bool (False)
            Ignore the cowding limit.

        Returns
        -------
        float
        The number of stars above the error limit
        """

        cols = [seeing_col, m5_col]
        units = "N stars"
        self.crowding_error = crowding_error
        self.m5_col = m5_col
        self.filtername = filtername
        self.seeing_col = seeing_col
        self.ignore_crowding = ignore_crowding
        if metric_name is None:
            metric_name = "N stars to Precision %.2f" % (crowding_error)
        super().__init__(col=cols, maps=maps, units=units, metric_name=metric_name, **kwargs)

    def run(self, data_slice, slice_point=None):
        pix_area = hp.nside2pixarea(slice_point["nside"], degrees=True)
        # Set mag_vector to the same length as starLumFunc (lower edge of mag bins)
        mag_vector = slice_point[f"starMapBins_{self.filtername}"][1:]
        # Pull up density of stars at this point in the sky
        lum_func = slice_point[f"starLumFunc_{self.filtername}"]
        # Calculate the crowding error using the best seeing value (in any filter?)
        crowd_error = _comp_crowd_error(mag_vector, lum_func, seeing=min(data_slice[self.seeing_col]))
        # Locate at which point crowding error is greater than user-defined limit
        above_crowd = np.where(crowd_error >= self.crowding_error)[0]

        if np.size(above_crowd) == 0:
            crowd_mag = max(mag_vector)
        else:
            crowd_mag = mag_vector[max(above_crowd[0] - 1, 0)]

        # Compute the coadded depth, and the mag where that depth hits the error specified
        coadded_depth = 1.25 * np.log10(np.sum(10.0 ** (0.8 * data_slice[self.m5_col])))
        mag_limit = -2.5 * np.log10(1.0 / (self.crowding_error * (1.09 * 5))) + coadded_depth

        # Use the shallower depth, crowding or coadded
        if self.ignore_crowding:
            min_mag = mag_limit
        else:
            min_mag = np.min([crowd_mag, mag_limit])

        # Interpolate to the number of stars
        result = (
            np.interp(
                min_mag,
                slice_point[f"starMapBins_{self.filtername}"][1:],
                slice_point[f"starLumFunc_{self.filtername}"],
            )
            * pix_area
        )

        return result


class CrowdingMagUncertMetric(BaseMetric):
    """
    Given a stellar magnitude, calculate the mean uncertainty on the magnitude from crowding.
    """

    def __init__(
        self,
        rmag=20.0,
        seeing_col="seeingFwhmGeom",
        units="mag",
        metric_name=None,
        filtername="r",
        maps=["StellarDensityMap"],
        **kwargs,
    ):
        """
        Parameters
        ----------
        rmag : float
            The magnitude of the star to consider.

        Returns
        -------
        float
            The uncertainty in magnitudes caused by crowding for a star of rmag.
        """

        self.filtername = filtername
        self.seeing_col = seeing_col
        self.rmag = rmag
        if metric_name is None:
            metric_name = "CrowdingError at %.2f" % (rmag)
        super().__init__(col=[seeing_col], maps=maps, units=units, metric_name=metric_name, **kwargs)

    def run(self, data_slice, slice_point=None):
        mag_vector = slice_point[f"starMapBins_{self.filtername}"][1:]
        lum_func = slice_point[f"starLumFunc_{self.filtername}"]
        # Magnitude uncertainty given crowding
        dmag_crowd = _comp_crowd_error(
            mag_vector, lum_func, data_slice[self.seeing_col], single_mag=self.rmag
        )
        result = np.mean(dmag_crowd)
        return result



__all__ = ("StellarDensityMap",)

import os

import healpy as hp
import numpy as np
from rubin_scheduler.data import get_data_dir

from rubin_sim.maf.utils import radec2pix

from . import BaseMap


class StellarDensityMap(BaseMap):
    """
    Return the cumulative stellar luminosity function for each slice_point. Units of stars per sq degree.
    Uses a healpix map of nside=64. Uses the nearest healpix point for other ra,dec values.

    Parameters
    ----------
    startype : `str` ('allstars', 'wdstars')
        Load the luminosity function for all stars ('allstars'), which includes main-sequence stars
        white dwarfs, blue horozontal branch, RR Lyrae, and Cepheids. The 'wdstars' option only includes
        white dwarf stars.
    filtername : `str`
        Filter to use. Options of u,g,r,i,z,y
    """

    def __init__(self, startype="allstars", filtername="r", map_dir=None):
        if map_dir is not None:
            self.map_dir = map_dir
        else:
            self.map_dir = os.path.join(get_data_dir(), "maps", "StarMaps")
        self.filtername = filtername
        self.keynames = [
            f"starLumFunc_{self.filtername}",
            f"starMapBins_{self.filtername}",
        ]
        if startype == "allstars":
            self.startype = ""
        else:
            self.startype = startype + "_"

    def _read_map(self):
        filename = "starDensity_%s_%snside_64.npz" % (self.filtername, self.startype)
        star_map = np.load(os.path.join(self.map_dir, filename))
        self.star_map = star_map["starDensity"].copy()
        self.star_map_bins = star_map["bins"].copy()
        self.starmap_nside = hp.npix2nside(np.size(self.star_map[:, 0]))

    def run(self, slice_points):
        self._read_map()

        nside_match = False
        if "nside" in slice_points:
            if slice_points["nside"] == self.starmap_nside:
                slice_points[f"starLumFunc_{self.filtername}"] = self.star_map
                nside_match = True
        if not nside_match:
            # Compute the healpix for each slice_point on the nside=64 grid
            indx = radec2pix(self.starmap_nside, slice_points["ra"], slice_points["dec"])
            slice_points[f"starLumFunc_{self.filtername}"] = self.star_map[indx, :]

        slice_points[f"starMapBins_{self.filtername}"] = self.star_map_bins
        return slice_points



__all__ = ("AreaSummaryMetric", "AreaThresholdMetric")

import healpy as hp
import numpy as np

from .base_metric import BaseMetric


class AreaSummaryMetric(BaseMetric):
    """
    Find the min/max of a value in the best area. This is a handy substitute for when
    users want to know "the WFD value".

    Parameters
    ----------
    area : float (18000)
        The area to consider (sq degrees)
    decreasing : bool (True)
        Should the values be sorted by increasing or decreasing order. For values where
        "larger is better", decreasing is probably what you want. For metrics where
        "smaller is better" (e.g., astrometric precission), set decreasing to False.
    reduce_func : None
        The function to reduce the clipped values by. Will default to min/max depending on
        the bool val of the decreasing kwarg.

    """

    def __init__(
        self,
        col="metricdata",
        metric_name="AreaSummary",
        area=18000.0,
        decreasing=True,
        reduce_func=None,
        **kwargs,
    ):
        super().__init__(col=col, metric_name=metric_name, **kwargs)
        self.area = area
        self.decreasing = decreasing
        self.reduce_func = reduce_func
        self.mask_val = np.nan  # Include so all values get passed
        self.col = col
        if reduce_func is None:
            if decreasing:
                self.reduce_func = np.min
            else:
                self.reduce_func = np.max
        else:
            self.reduce_func = reduce_func

    def run(self, data_slice, slice_point=None):
        # find out what nside we have
        nside = hp.npix2nside(data_slice.size)
        pix_area = hp.nside2pixarea(nside, degrees=True)
        n_pix_needed = int(np.ceil(self.area / pix_area))

        # Only use the finite data
        data = data_slice[self.col][np.isfinite(data_slice[self.col].astype(float))]
        order = np.argsort(data)
        if self.decreasing:
            order = order[::-1]
        result = self.reduce_func(data[order][0:n_pix_needed])
        return result


class AreaThresholdMetric(BaseMetric):
    """
    Find the amount of area on the sky that meets a given threshold value.

    The area per pixel is determined from the size of the metric_values array passed to the summary metric.
    This assumes that both all values are passed and that the metric was calculated with a healpix slicer.

    Parameters
    ----------
    upper_threshold : `float` or None
        The metric value must be below this threshold to count toward the area.
        Default None implies no upper bound.
    lower_threshold : `float` or None, opt
        The metric value must be above this threshold to count toward the area.
        Default None implies no lower bound.
    """

    def __init__(
        self,
        col="metricdata",
        metric_name="AreaThreshold",
        upper_threshold=None,
        lower_threshold=None,
        **kwargs,
    ):
        super().__init__(col=col, metric_name=metric_name, **kwargs)
        self.upper_threshold = upper_threshold
        self.lower_threshold = lower_threshold
        self.mask_val = np.nan  # Include so all values get passed
        self.col = col
        self.units = "degrees"

    def run(self, data_slice, slice_point=None):
        # find out what nside we have
        nside = hp.npix2nside(data_slice.size)
        pix_area = hp.nside2pixarea(nside, degrees=True)
        # Look for pixels which match the critera for the thresholds
        if self.upper_threshold is None and self.lower_threshold is None:
            npix = len(data_slice)
        elif self.upper_threshold is None:
            npix = len(np.where(data_slice[self.col] > self.lower_threshold)[0])
        elif self.lower_threshold is None:
            npix = len(np.where(data_slice[self.col] < self.upper_threshold)[0])
        else:
            npix = len(
                np.where(
                    (data_slice[self.col] > self.lower_threshold)
                    and (data_slice[self.col] < self.upper_threshold)
                )[0]
            )
        area = pix_area * npix
        return area
